"""
File: ut_softwareupdater_localtest.py
"""

#pragma out [ PASS ]

import os
import sys
import imp
import glob
import shutil
import logging
import urllib2
import StringIO
import tempfile
import subprocess
import multiprocessing
from SocketServer import TCPServer
from SimpleHTTPServer import SimpleHTTPRequestHandler


def _repo_server(server_root, port, lockfile_path):
	sys.stdout = StringIO.StringIO()
	sys.stderr = StringIO.StringIO()
	os.chdir(server_root)
	srvr = TCPServer(("localhost", int(port)), SimpleHTTPRequestHandler)
	while os.path.exists(lockfile_path):
		srvr.handle_request()



class Scenario:

	def __init__(self, project_root, port=12345):

		self.port = port
		self.address = None
		self.server = None

		self.lockfile_path = None

		self.project_root = project_root
		self.scenario_root = tempfile.mkdtemp()

		self._prepare_scenario_root()

		self.client_root = os.path.join(self.scenario_root, "client")

		self.client_repo_path = os.path.join(self.client_root, "repo")

		self.client_current_metadata = os.path.join(self.client_repo_path, "cur")

		self.client_previous_metadata = os.path.join(self.client_repo_path, "prev")

		self.server_root = os.path.join(self.scenario_root, "server")
		os.mkdir(self.server_root)

		self.server_metadata_path = os.path.join(self.server_root, "meta")
		os.mkdir(self.server_metadata_path)

		self.server_targets_path = os.path.join(self.server_root, "targets")

		self.keystore_path = os.path.join(self.scenario_root, "keystore")

	def _prepare_scenario_root(self):
		preparetest_args = [	"python",
					"preparetest.py",
					"-t",
					self.scenario_root]
		subprocess.call(preparetest_args, 
				stdout=subprocess.PIPE,
				stderr=subprocess.PIPE)
		sys.path.append(self.scenario_root)
		import softwareupdater
		self.updater = softwareupdater

	def run_test(self):
		self.pre_server_action()
		self.make_server()
		self.post_server_action()
		self.make_client()
		self.post_client_action()
		tmp_dir = tempfile.mkdtemp()
		updates = self.updater.do_rsync(	self.address,
							self.client_root,
							tmp_dir,
							self.client_root)
		return self.evaluate_updates(updates)


	def make_client(self):
		shutil.copytree(self.project_root, self.client_root)
		os.makedirs(self.client_current_metadata)
		os.makedirs(self.client_previous_metadata)
		for f in glob.iglob(os.path.join(self.server_metadata_path, "*")):
			shutil.copy(f, self.client_current_metadata)
		for f in glob.iglob(os.path.join(self.server_metadata_path, "*")):
			shutil.copy(f, self.client_previous_metadata)

	def make_server(self):
		args = ["quickstart.py",
			"-k",
			self.keystore_path,
			"-p",
			"password",
			"-t",
			"1",
			"-l",
			self.server_root,
			"-r",
			self.project_root,
			"-e",
			"12/12/2012"]
		subprocess.call(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		self.start_repo_server()

	def start_repo_server(self):
		self.lockfile_path = tempfile.mktemp()
		open(self.lockfile_path, "w").close()
		srvr_args = (self.server_root, self.port, self.lockfile_path)
		self.server = multiprocessing.Process(	target=_repo_server, 
							args=srvr_args)
		self.server.daemon = True
		self.server.start()
		self.address = "http://localhost:%s" % self.port

	def stop_repo_server(self):
		if self.server is None:
			return
		if self.address is None:
			raise Exception("Address set but not the server!")
		if self.lockfile_path:
			os.remove(self.lockfile_path)

		try: urllib2.urlopen(self.address)
		except Exception: pass

		self.server.join()

	def pre_server_action(self):
		pass

	def post_server_action(self):
		pass

	def post_client_action(self):
		pass

	def evaluate_updates(self, updates):
		pass

	def cleanup(self):
		self.stop_repo_server()
		shutil.rmtree(self.scenario_root)



class NoUpdateTest(Scenario):

	def evaluate_updates(self, updates):
		if updates: return False
		return True


class BadSignatureTest(Scenario):

	def pre_server_action(self):
		# we actually build a separate server to connect to
		self.noup = NoUpdateTest(self.project_root, 12346)
		
	def post_client_action(self):
		# now we tell the client to connect to that server
		self.port = 12346

	def cleanup(self):
		self.stop_repo_server()
		self.noup.stop_repo_server()
		shutil.rmtree(self.scenario_root)


class UpdateTest(Scenario):

	def post_client_action(self):
		# stop the existing server
		self.stop_repo_server()
		self.port = 12346
		self.address = "http://localhost:12346"
		cfg_path = os.path.join(self.server_root, 'root.cfg')
		shutil.copy(cfg_path, self.project_root)
		shutil.rmtree(self.server_targets_path)
		shutil.copytree(self.client_root, self.server_targets_path)
		args = ["quickstart.py",
			"-u",
			"-k",
			self.keystore_path,
			"-p",
			"password",
			"-l",
			self.server_root,
			"-r",
			self.server_targets_path,
			"-c",
			cfg_path]
		subprocess.call(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		self.start_repo_server()
		os.remove(cfg_path)

	def evaluate_updates(self, updates):
		correct_updates = [	'repo/cur/targets.txt', 
					'repo/cur/root.txt', 
					'repo/cur/timestamp.txt',
					'repo/cur/release.txt']
		if set(updates) != set(correct_updates):
			return False
		return True

if __name__ == "__main__":

	results = []

	noup = NoUpdateTest(".")
	results.append(noup.run_test())
	noup.cleanup()
	
	bad_sig = BadSignatureTest(".")
	try:
		err = False
		bad_sig.run_test()
	except Exception:
		err = True
	results.append(err)
	bad_sig.cleanup()

	fourup = UpdateTest(".")
	results.append(fourup.run_test())
	fourup.cleanup()

	if all(results) and (len(results) == 3):
		print "[ PASS ]"
	else:
		print "[ FAIL ]"
